%%% -*- erlang-indent-level: 2 -*-
%%%-------------------------------------------------------------------
%%% File    : dialyzer_cl.erl
%%% Authors : Tobias Lindahl <tobiasl@csd.uu.se>
%%%           Kostis Sagonas <kostis@it.uu.se>
%%% Description : The command line interface for the Dialyzer tool.
%%%
%%% Created : 27 Apr 2004 by Tobias Lindahl <tobiasl@csd.uu.se>
%%%-------------------------------------------------------------------
-module(dialyzer_cl).

-export([start/1, check_init_plt/0]).

-include("dialyzer.hrl").	  %% file is automatically generated
-include("hipe_icode_type.hrl").

-record(cl_state, {backend_pid, legal_warnings=[], 
		   output=standard_io, user_plt,
		   nof_warnings=0, return_status=0}).

start(Options) when is_list(Options) ->
  start(dialyzer_options:build(Options));
start(#options{} = DialyzerOptions) ->
  process_flag(trap_exit, true),
  State = new_state(),
  NewState1 = init_output(State, DialyzerOptions),
  NewState2 = NewState1#cl_state{legal_warnings=
				 DialyzerOptions#options.legal_warnings},
  InitAnalysis = build_analysis_record(NewState2, DialyzerOptions),
  NewState3 = run_analysis(NewState2, InitAnalysis),
  cl_loop(NewState3).


check_init_plt() ->
  process_flag(trap_exit, true),
  Plt = filename:join(os:getenv("DIALYZER_DIR"), "plt/dialyzer_init_plt"),
  ReAnalyze = 
    case filelib:is_file(Plt) of
      true ->
	case dets:open_file(Plt) of
	  {ok, Dets} ->
	    Res =
	      case dets:lookup(Dets, md5) of
		{error, _} -> true;
		[] -> true;
		[{md5, OldMD5}] ->
		  case compute_md5() of
		  OldMD5 -> false;
		    Other ->  {md5, Other}
		  end
	      end,
	    ok = dets:close(Dets),
	    Res;
	  {error, _} ->
	    true
	end;
      false -> true
    end,
  Msg = "    Creating one (will take a few minutes)\n",
  case ReAnalyze of
    true ->
      io:format(" no\n~s", [Msg]),
      create_init_plt();
    {md5, MD5} ->
      io:format(" no\n~s", [Msg]),
      create_init_plt(MD5);
    false ->
      io:format(" yes\n"),
      ok
  end.

compute_md5() ->
  LibDir = code:lib_dir("stdlib"),
  Dir = filename:join(LibDir, "ebin"),
  {ok,List} = file:list_dir(Dir),
  BeamFiles = [filename:join(Dir, X) 
	       || X<-List, filename:extension(X)==".beam"],
  Context = erlang:md5_init(),
  Fun =
    fun(X, Acc)->
	{ok,Bin} = file:read_file(X),
	erlang:md5_update(Acc, Bin)
    end,
  FinalContext = lists:foldl(Fun, Context, BeamFiles),
  erlang:md5_final(FinalContext).

create_init_plt() ->
  MD5 = compute_md5(),
  create_init_plt(MD5).

create_init_plt(MD5) ->  
  State = new_state(),
  EmptyPLT = filename:join(os:getenv("DIALYZER_DIR"), "plt/dialyzer_empty_plt"),
  PLT = filename:join(os:getenv("DIALYZER_DIR"), "plt/dialyzer_init_plt"),
  Files = [filename:join(code:lib_dir("stdlib"),"src")],
  Includes = [filename:join(code:lib_dir(kernel), "include")],
  Analysis = #analysis{fixpoint=first, user_plt = PLT, files=Files, 
		       granularity=all, init_plt=EmptyPLT,
		       include_dirs=Includes, md5=MD5,
		       start_from=src_code,
		       check_inline=true},
  cl_loop(run_analysis(State, Analysis)).

new_state() ->
  User =
    case os:getenv("USER") of
      false ->      
	os:cmd("whoami");
      U -> U
    end,
  UserPLT = filename:join(os:getenv("DIALYZER_TMP"), "dialyzer_plt." ++ User),
		  
  #cl_state{user_plt=UserPLT}.

init_output(State, DialyzerOptions) ->
  case DialyzerOptions#options.output_file of
    "" ->
      State;
    OutputFile ->
      case file:open(OutputFile, [write]) of
	{ok, File} ->
	  State#cl_state{output=File};
	{error, Reason} ->
	  io:format("Could not open output file ~p, Reason ~p\n",
		    [OutputFile, Reason]),
	  exit(error)
      end
  end.

maybe_close_output_file(State) ->
  case State#cl_state.output of
    standard_io -> ok;
    File -> file:close(File)
  end.

			 
%% ----------------------------------------------------------------
%%
%%  Main Loop
%%

cl_loop(State) ->
  BackendPid = State#cl_state.backend_pid,
  Output = State#cl_state.output,
  receive
    {BackendPid, log, _LogMsg} ->
      %% io:format(Output,"Log: ~s", [LogMsg]),
      cl_loop(State);
    {BackendPid, warnings, Warnings} ->
      NofOldWarnings = State#cl_state.nof_warnings,
      case NofOldWarnings of
	0 -> io:format(Output, "\n", []); %% warnings are just starting to appaer
	_ -> ok
      end,
      io:format(Output, "~s", [Warnings]),
      NofNewWarning = length(string:tokens(lists:flatten(Warnings), "\n")),
      NewState = State#cl_state{nof_warnings=NofOldWarnings+NofNewWarning},
      cl_loop(NewState);
    {BackendPid, error, Msg} ->
      io:format(Output, "~s", [Msg]),
      cl_loop(State#cl_state{return_status=-1});
    {BackendPid, done} ->
      return_value(State);
    {'EXIT', BackendPid, {error, Reason}} ->
      Msg = io_lib:format("Analysis failed with error report: ~s\n", 
		          [Reason]),
      error(State, Msg);
    {'EXIT', BackendPid, Reason} when Reason /= normal ->
      Msg = io_lib:format("Analysis failed with error report: ~p\n", 
		          [Reason]),
      maybe_close_output_file(State),
      error(State, Msg);
    _Other ->
      %% io:format(Output, "Received ~p\n", [Other]),
      cl_loop(State)
  end.

error(State, Msg) ->
  io:format(State#cl_state.output, "~s", [Msg]),
  return_value(State#cl_state{return_status=-1}).

return_value(State = #cl_state{return_status=-1}) ->
  maybe_close_output_file(State),
  ?RET_INTERNAL_ERROR;
return_value(State = #cl_state{nof_warnings=NofWarnings}) ->
  maybe_close_output_file(State),
  if NofWarnings == 0 ->
      ?RET_NOTHING_SUSPICIOUS;
     true ->
      ?RET_DISCREPANCIES_FOUND
  end.
  

%% ---- PLT ----

init_plt(From, To) ->
  case file:copy(From, To) of
    {ok, _} ->
      ok;
    {error, Reason} ->      
      {error, io_lib:format("Could not init lookup table: ~w",[Reason])}
  end.

%% ----------------------------------------------------------------
%%
%%  Run the analysis
%%

build_analysis_record(State, DialyzerOptions) ->
  PLT = State#cl_state.user_plt,
  InitPLT = filename:join(os:getenv("DIALYZER_DIR"), "plt/dialyzer_init_plt"),
  From = DialyzerOptions#options.from,
  IncludeDirs = DialyzerOptions#options.include_dirs,
  Checkinline = From == byte_code,
  Defines = DialyzerOptions#options.defines,
  Files1 = ordsets:from_list(DialyzerOptions#options.files),
  Files2 = add_files_rec(DialyzerOptions#options.files_rec, From),  
  Files = ordsets:union(Files1, Files2),

  #analysis{fixpoint=first, defines=Defines, granularity=all,
	    include_dirs=IncludeDirs, init_plt=InitPLT, user_plt=PLT, 
	    files=Files, start_from=From, check_inline=Checkinline}.


add_files_rec(Files, From) ->
  Files1 = ordsets:from_list(Files), 
  Dirs1 = ordsets:filter(fun(X) -> filelib:is_dir(X)end, Files1),  
  Dirs2 = ordsets:union(Dirs1, all_subdirs(Dirs1)),  
  FinalFiles = ordsets:union(Files1, Dirs2),
  case From of
    byte_code -> filter_files(FinalFiles, ".beam");
    src_code -> filter_files(FinalFiles, ".erl")
  end.

all_subdirs(Dirs) ->
  all_subdirs(Dirs, []).

all_subdirs([Dir|T], Acc) ->
  {ok, Files} = file:list_dir(Dir),
  SubDirs = lists:zf(fun(F) ->
                       SubDir = filename:join(Dir, F),
                       case filelib:is_dir(SubDir) of
                         true -> {true, SubDir};
                         false -> false
                       end
                   end, Files),
  NewAcc = ordsets:union(ordsets:from_list(SubDirs), Acc),
  all_subdirs(T++SubDirs, NewAcc);
all_subdirs([], Acc) ->
  Acc.

filter_files(Files, Extension) ->
  Fun = fun(X) -> 
	    filename:extension(X)==Extension 
	      orelse 
		(filelib:is_dir(X) andalso
		 contains_files(X, Extension))
	end,
  lists:filter(Fun, Files).

contains_files(Dir, Extension) ->
  {ok, Files} = file:list_dir(Dir),
  lists:any(fun(X)->filename:extension(X)==Extension end,Files).


run_analysis(State, Analysis) ->
  case init_plt(Analysis#analysis.init_plt, Analysis#analysis.user_plt) of
    ok -> ok;
    {error, Reason} ->
      io:format(State#cl_state.output, "~s", [Reason]),
      exit(error)
  end,
  Self = self(),
  LegalWarnings = State#cl_state.legal_warnings,
  Fun = fun() -> dialyzer_analysis:start(Self, LegalWarnings, Analysis) end,
  BackendPid = spawn_link(Fun),
  State#cl_state{backend_pid=BackendPid}.
