%%%-------------------------------------------------------------------
%%% File    : dialyzer_plt.erl
%%% Author  : Tobias Lindahl <tobiasl@it.uu.se>
%%% Description : Interface to display information in the persistent 
%%%               lookup tables.
%%%
%%% Created : 23 Jul 2004 by Tobias Lindahl <tobiasl@it.uu.se>
%%%-------------------------------------------------------------------
-module(dialyzer_plt).

-export([
	 to_edoc/1,
	 to_edoc/4
	]).

to_edoc(PLT) ->
  to_edoc(PLT, '_', '_', '_').

to_edoc(PLT, M, F, A) ->
  case plt_type(PLT) of
    ets ->      
      List = ets:match_object(PLT, {{M, F, A}, '_'}),
      SortedList = lists:keysort(1, List),
      ExpandedList = lists:flatten(expand_edoc(SortedList, [])),
      {ok, ExpandedList};
    dets ->
      case open_dets(PLT) of
	{ok, Dets} ->
	  List = dets:match_object(Dets, {{M, F, A}, '_'}),
	  SortedList = lists:keysort(1, List),
	  ExpandedList = lists:flatten(expand_edoc(SortedList, [])),
	  close_dets(Dets),
	  {ok, ExpandedList};
	{error, _Reason} = Return ->
	  Return
      end;
    {error, _} = Return -> 
      Return
  end.

plt_type(PLT) ->
  case lists:member(PLT, ets:all()) of
    true -> 
      ets;
    false ->
      case dets:is_dets_file(PLT) of
	true ->
	  dets;
	false ->
	  {error, io_lib:format("Could not open documentation plt: ~w", [PLT])};
	{error, _} ->
	  {error, io_lib:format("Could not open documentation plt: ~w", [PLT])}
      end
  end.

expand_edoc([{{M, F, _A}, {ArgTypes, ReturnType}}| Left], M) ->
  [io_lib:format("%% @spec ~w(~s) -> ~s\n", [F, expand_args(ArgTypes), 
					     erl_types:t_to_string(ReturnType)])
   | expand_edoc(Left, M)];
expand_edoc([{{M, F, A}, ReturnType}| Left], M) ->
  [io_lib:format("%% @spec ~w(~s) -> ~s\n", [F, expand_args_any(A), 
					     erl_types:t_to_string(ReturnType)])
   | expand_edoc(Left, M)];
expand_edoc(List = [{{M1, _F, _A}, _Type}| _], _M) ->
  [io_lib:format("\n\n%% -------  Module: ~w -------\n\n", [M1]) | 
   expand_edoc(List, M1)];
expand_edoc([], _) ->
  [].

expand_args_any(0) ->
  [];
expand_args_any(1) ->
  [erl_types:t_to_string(erl_types:t_any())];
expand_args_any(X) ->
  [erl_types:t_to_string(erl_types:t_any()) ++ ","|expand_args_any(X-1)].

expand_args([]) ->
  [];
expand_args([ArgType]) ->
  [erl_types:t_to_string(ArgType)];
expand_args([ArgType|Left]) ->
  [erl_types:t_to_string(ArgType) ++ ","|expand_args(Left)].

open_dets(File) ->
  case dets:open_file(File, []) of
    {ok, _Dets} = Return -> 
      Return;
    {error, Reason} ->
      {error, io_lib:format("Could not open plt: ~w", [Reason])}
  end.

close_dets(File) ->
  dets:close(File).
