%% -*- erlang-indent-level: 2 -*-
%%====================================================================
%% File        : dialyzer_chk.erl
%% Author      : Kostis Sagonas <kostis@it.uu.se>
%% Description : A consistency checker for Dialyzer.
%% Created     : 27 Nov 2004
%% 
%% $Id: dialyzer_chk.erl,v 1.2 2004/12/12 10:15:12 kostis Exp $
%%====================================================================

-module(dialyzer_chk).

-export([report_to_file/0, test/1]).

report_to_file() ->
  %% The following relies on this function being executed in
  %% Dialyzer's main dir.
  {ok,DialyzerDir} = file:get_cwd(),
  Res = test(DialyzerDir),
  {ok,File} = file:open(atom_to_list(?MODULE) ++ "_new", [write]),
  io:fwrite(File, "~p\n", [Res]).

test([DialyzerDir]) when atom(DialyzerDir) ->	% this clause handles calls
  test(atom_to_list(DialyzerDir));		% from the command line
test(DialyzerDir) ->
  Apps = ["compiler",
          "kernel",
	  "stdlib",
	  "hipe",
	  "gs"],                                % gs is needed by the gui
  Path = [filename:join(code:lib_dir(App), "ebin") || App <- Apps],
  Server = dialyzer_testsuite,
  xref:start(Server),
  xref:set_default(Server, [{verbose,false},{warnings,false}]),
  xref:set_library_path(Server, Path),
  {ok,_App} = xref:add_application(Server, DialyzerDir),
  {ok,Undef} = xref:analyze(Server, undefined_function_calls),
  {ok,UnusedLocals} = xref:analyze(Server, locals_not_used),
  {ok,UnusedExports} = xref:analyze(Server, exports_not_used),
  ReallyUnusedExports = lists:subtract(UnusedExports, used_exports()),
  catch xref:stop(Server),
  U1 = case Undef of
	 [] -> no_undef_funs;
	 _ ->
	   io:format("\n"),
	   lists:foreach(fun ({MFA1,MFA2}) ->
			     io:format("~s calls undefined ~s\n",
				       [format_mfa(MFA1),
					format_mfa(MFA2)])
			 end, Undef),
	   {undefined_functions_in_dialyzer,Undef}
       end,
  U2 = case UnusedLocals of
	 [] -> no_unused_local_funs;
	 _ ->
	   io:format("\n"),
	   lists:foreach(fun (MFA) ->
			     io:format("unused: ~s\n", [format_mfa(MFA)])
			 end, UnusedLocals),
	   {unused_locals_in_dialyzer,UnusedLocals}
       end,
  U3 = case ReallyUnusedExports of
	 [] -> no_unused_exported_funs;
	 _ ->
	   io:format("\n"),
	   lists:foreach(fun (MFA) ->
			     io:format("exported but unused: ~s\n",
				       [format_mfa(MFA)])
			 end, ReallyUnusedExports),
	   {unused_exports_in_dialyzer,ReallyUnusedExports}
       end,
  case {U1,U2,U3} of
    {no_undef_funs,no_unused_local_funs,no_unused_exported_funs} ->
      io:format("ok\n"),
      ok;
    Other ->
      Other
  end.

format_mfa({M,F,A}) ->
  io_lib:format("~s:~s/~p", [M,F,A]).

%%=====================================================================
%% Below appears some hard-coded information about exported functions
%% which are used either outside the Dialyzer application.
%%=====================================================================

used_exports() ->
  application_interface().

%%
%% This is the list of all functions which are supposed to be
%% accessible by the user or by other applications.
%%
application_interface() ->
  [
   {dialyzer,cl,1},
   {dialyzer,gui,1},
   {dialyzer,run,1},
   {dialyzer,cl_check_init_plt,0}
  ].

